"""
    repfactors = replication_factors(unitcell, cutoffradius)
Find the replication factors needed to make a supercell big enough to fit a sphere with the specified cutoff radius.
In Xtals.jl, rather than replicating the atoms in the home unit cell to build the supercell that
serves as a simulation box, we replicate the home unit cell to form the supercell (simulation box) in a for loop.
This function ensures enough replication factors such that the nearest image convention can be applied.
A non-replicated supercell has 1 as the replication factor in each dimension (`repfactors = (1, 1, 1)`).
# Arguments
- `unitcell::Box`: The unit cell of the framework
- `cutoff_radius::Float64`: Cutoff radius beyond which we define the potential energy to be zero (units: Angstrom)
# Returns
- `repfactors::Tuple{Int, Int, Int}`: The replication factors in the a, b, c directions
"""
function replication_factors(unitcell::Box, cutoff_radius::Float64)
    # Unit vectors used to transform from fractional coordinates to cartesian coordinates. We'll be
    a = unitcell.f_to_c[:, 1]
    b = unitcell.f_to_c[:, 2]
    c = unitcell.f_to_c[:, 3]

    n_ab = cross(a, b)
    n_ac = cross(a, c)
    n_bc = cross(b, c)

    # c0 defines a center in the unit cell
    c0 = [a b c] * [.5, .5, .5]

    rep = [1, 1, 1]

    # Repeat for `a`
    # |n_bc ⋅ c0|/|n_bc| defines the distance from the end of the supercell and the center. As long as that distance is less than the cutoff radius, we need to increase it
    while abs(dot(n_bc, c0)) / norm(n_bc) < cutoff_radius
        rep[1] += 1
        a += unitcell.f_to_c[:,1]
        c0 = [a b c] * [.5, .5, .5]
    end

    # Repeat for `b`
    while abs(dot(n_ac, c0)) / norm(n_ac) < cutoff_radius
        rep[2] += 1
        b += unitcell.f_to_c[:,2]
        c0 = [a b c] * [.5, .5, .5]
    end

    # Repeat for `c`
    while abs(dot(n_ab, c0)) / norm(n_ab) < cutoff_radius
        rep[3] += 1
        c += unitcell.f_to_c[:,3]
        c0 = [a b c] * [.5, .5, .5]
    end

    return (rep[1], rep[2], rep[3])::Tuple{Int, Int, Int}
end
